/*
 * Implementation of class StillImage
 *
 * $Date: 2004/05/21 09:22:58 $
 * $Revision: 1.8 $
 *
 */

#include "WImage/ImageArray.hh"
#include "WImage/StillImage.hh"
#include "WImage/PgmReader.hh"
#include "WImage/RawReader.hh"
#ifdef PFI
#include "WImage/PfgReader.hh"
#include "WImage/PfgWriter.hh"
#endif
#include "WImage/PgmWriter.hh"
#include "WImage/RawWriter.hh"
#include <stdexcept>
#include <iostream>
#include <string.h>
#include <assert.h>

#ifndef __WATCOMC__
using namespace std;
#endif

Image *
StillImage::mkImage (int rows, int cols)
{
  return new StillImage (rows, cols);
}

StillImage::StillImage (int rows, int cols): Image ()
{
  //cout << "StillImage: " << rows << " " << cols << endl;
  m_ysize = rows;
  m_xsize = cols;
  m_xysize = rows * cols;
  m_coeffs = new ImageArray<coeff> (rows, cols);
  m_acol = new ColumnVector (m_coeffs);
  m_arow = new RowVector (m_coeffs);
  m_afull = new FullVector (m_coeffs);
  m_asteps = m_ssteps = 0;
}

StillImage::StillImage (ImageArray<coeff> &ia): Image ()
{
  m_ysize = ia.rows ();
  m_xsize = ia.cols ();
  m_xysize = ia.size ();
  m_coeffs = ia.clone ();
  m_acol = new ColumnVector (m_coeffs);
  m_arow = new RowVector (m_coeffs);
  m_afull = new FullVector (m_coeffs);
  m_asteps = m_ssteps = 0;
}

StillImage::~StillImage (void)
{
  delete m_coeffs;
  delete m_acol;
  delete m_arow;
  delete m_afull;
}

bool 
StillImage::epsilons (Image &si, coeff epsilon) const
{
  return ((StillImage*)this)->m_coeffs->epsilons (((StillImage*)&si)->m_coeffs,
                                                  epsilon); 
}

bool 
StillImage::equals (Image &si) const
{ 
  return ((StillImage*)this)->m_coeffs->epsilons (((StillImage*)&si)->m_coeffs,
                                                  COEFF_EPSILON); 
}

void 
StillImage::read (char const * fname, int rawy, int rawx)
{
  FileName fn (fname);
  read (fname, fn.guess (), rawy, rawx);
  m_acol->update ();
  m_arow->update ();
  m_afull->update ();
}

void
StillImage::read (char const * fname, filetype ftype, int rawy, int rawx)
{
  GreymapReader *ir;
  switch (ftype)
    {
    case fn_pgm:
      ir = new PgmReader (fname, *m_coeffs);
      break;
    case fn_raw:
      if (rawy > 0 || rawx > 0)
        {
          ir = new RawReader (fname, *m_coeffs, rawy, rawx);
        }
      else
        {
          ir = new RawReader (fname, *m_coeffs);
        }
      break;
#ifdef PFI
    case fn_pfi:
      ir = new PfgReader (fname, *m_coeffs);
      break;
#endif
    default:
      throw invalid_argument ("read: unknown file type.");
      break;
    }
  ir->read ();
  m_ysize = m_coeffs->rows ();
  m_xsize = m_coeffs->cols ();
  m_xysize = m_coeffs->size ();
}

void 
StillImage::write (char const *fname, bool beautify)
{
  FileName fn (fname);
  write (fname, fn.guess (), beautify);
}

void
StillImage::write (char const *fname, filetype ftype, bool beautify)
{
  GreymapWriter *iw;
  bool ispixel = true;
  
  switch (ftype)
    {
    case fn_pgm:
      iw = new PgmWriter (fname, *m_coeffs);
      break;
    case fn_raw:
      iw = new RawWriter (fname, *m_coeffs);
      break;
#ifdef PFI
    case fn_pfi:
      iw = new PfgWriter (fname, *m_coeffs);
      ispixel = false;
      break;
#endif
    default:
      throw invalid_argument ("write: unknown file type.");
      break;
    }

  ImageArray<coeff> *backup = m_coeffs->clone ();

  if (beautify 
      || (ispixel && (m_asteps - m_ssteps > 0 || (this->smin () < 0))))
    {
      this->beautify ();
    }

  iw->write ();

  delete m_coeffs;
  m_coeffs = backup;

  delete m_acol;
  delete m_arow;
  delete m_afull;
  
  m_acol = new ColumnVector (m_coeffs);
  m_arow = new RowVector (m_coeffs);
  m_afull = new FullVector (m_coeffs);
}

Image *
StillImage::clone (void) const
{
  Image *ret = new StillImage (*m_coeffs);
  ret->anasteps (this->m_asteps);
  ret->synsteps (this->m_ssteps);
  return ret;
}

void 
StillImage::resize (int rows, int cols)
{
  m_ysize = rows;
  m_xsize = cols;
  m_xysize = rows * cols;
  m_coeffs->resize (rows, cols);
  m_acol->update ();
  m_arow->update ();
  m_afull->update ();
}

